"connection package"
__version__ = "0.3.0"

from snowflake.connector import DatabaseError, InterfaceError, ProgrammingError
from snowflake.connector.cursor import ResultMetadata

from .conn import Connection, Cursor, conn_opts, connection_names, getconn
from .jwt import get_token
from .utils import with_connection, with_connection_args, with_connection_options

__all__ = [
    "conn_opts",
    "getconn",
    "connection_names",
    "get_token",
    "Connection",
    "ResultMetadata",
    "Cursor",
    "DatabaseError",
    "ProgrammingError",
    "InterfaceError",
    "with_connection_args",
    "with_connection",
    "with_connection_options",
]



"test decorators"
from argparse import ArgumentParser
from dataclasses import dataclass
from typing import Any

from sfconn import Connection, with_connection, with_connection_args


def test_decorators() -> None:
    "test decorators"

    @dataclass
    class Result:
        user: str
        role: str

    @with_connection_args(__doc__)
    def getargs(_: ArgumentParser) -> Any:
        pass

    @with_connection
    def main(cnx: Connection) -> None:
        with cnx.cursor() as csr:
            rows = list(csr.run_query(Result, "select current_user() as user, current_role() as role"))
            assert csr.rowcount == 1
            assert isinstance(rows[0], Result)

    main(**vars(getargs([])))  # type: ignore



"test connection options"
from pathlib import Path

import pytest

from sfconn import get_token


def test_keypair(config_keypair: Path) -> None:
    assert get_token("default") is not None


def test_non_keypair(config_password: Path) -> None:
    with pytest.raises(ValueError):
        get_token("default")



"get a JWT token"
import base64
import datetime as dt
import hashlib
from pathlib import Path
from typing import cast

import jwt

from .conn import conn_opts
from .privkey import PrivateKey

LIFETIME = dt.timedelta(minutes=59)  # The tokens will have a 59 minute lifetime
RENEWAL_DELTA = dt.timedelta(minutes=54)  # Tokens will be renewed after 54 minutes
ALGORITHM = "RS256"  # Tokens will be generated using RSA with SHA256


def fingerprint(pubkey: bytes) -> str:
    "base64 encoded fingerprint of the public key"
    sha256hash = hashlib.sha256()
    sha256hash.update(pubkey)
    return "SHA256:" + base64.b64encode(sha256hash.digest()).decode("utf-8")


def _clean_account_name(account: str) -> str:
    "ref: https://docs.snowflake.com/en/developer-guide/sql-api/authenticating.html#generating-a-jwt-in-python"
    if ".global" not in account:
        if (idx := account.find(".")) > 0:
            return account[:idx]
    else:
        if (idx := account.find("-")) > 0:
            return account[:idx]
    return account


def get_token(connection_name: str | None = None, lifetime: dt.timedelta = LIFETIME) -> str:
    """get a JWT when using key-pair authentication

    Args
        conn: A connection name to be looked up from the config_file, optional, default to None for the default connection
        lifetime: issued token's lifetime

    Returns:
        a JWT

    Exceptions:
        ValueError: if `conn` doesn't support key-pair authentication
        *: Any exceptions raised by either conn_opts() or class PrivateKey
    """

    opts = conn_opts(connection_name)
    keyf = cast(str | None, opts.get("private_key_file"))
    if keyf is None:
        raise ValueError(f"'{connection_name}' does not use key-pair authentication to support creating a JWT")

    qual_user = f"{_clean_account_name(opts['account']).upper()}.{opts['user'].upper()}"

    key = PrivateKey(Path(keyf))
    now = dt.datetime.now()

    payload = {
        "iss": f"{qual_user}.{fingerprint(key.pub_bytes)}",
        "sub": f"{qual_user}",
        "iat": int(now.timestamp()),
        "exp": int((now + lifetime).timestamp()),
    }

    return jwt.encode(payload, key=key.key, algorithm=ALGORITHM)  # type: ignore



"Utility functions"
import logging
from argparse import SUPPRESS, ArgumentParser, ArgumentTypeError
from functools import wraps
from pathlib import Path
from typing import Any, Callable

from .conn import conn_opts, getconn

_loglevel = logging.WARNING


def init_logging(logger: logging.Logger) -> None:
    "initialize the logging system"
    h = logging.StreamHandler()
    h.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    logger.addHandler(h)
    logger.setLevel(_loglevel)


def with_connection_options(
    fl: Callable[..., None] | logging.Logger | None = None,
) -> Callable[..., None] | Callable[[Callable[..., None]], Callable[..., None]]:
    "wraps application entry function that expects a connection"

    logger: logging.Logger | None = None

    def wrapper(fn: Callable[..., None]) -> Callable[..., None]:
        @wraps(fn)
        def wrapped(
            keyfile_pfx_map: tuple[Path, Path] | None,
            connection_name: str | None,
            database: str | None,
            role: str | None,
            schema: str | None,
            warehouse: str | None,
            loglevel: int,
            **kwargs: Any,
        ) -> None:
            "script entry-point"
            global _loglevel

            _loglevel = loglevel
            init_logging(logging.getLogger(__name__))
            if logger is not None:
                init_logging(logger)
            _opts = conn_opts(
                connection_name,
                keyfile_pfx_map=keyfile_pfx_map,
                database=database,
                role=role,
                schema=schema,
                warehouse=warehouse,
            )
            return fn(_opts, **kwargs)

        return wrapped

    if fl is None or isinstance(fl, logging.Logger):
        logger = fl
        return wrapper

    return wrapper(fl)


def with_connection(
    fl: Callable[..., None] | logging.Logger | None = None,
) -> Callable[..., None] | Callable[[Callable[..., None]], Callable[..., None]]:
    "wraps application entry function that expects a connection"

    logger = fl if isinstance(fl, logging.Logger) else None

    def wrapper(fn: Callable[..., None]) -> Callable[..., None]:
        @with_connection_options(logger)
        @wraps(fn)
        def wrapped(opts: dict[str, Any], **kwargs: Any) -> None:
            "script entry-point"
            try:
                with getconn(**opts) as cnx:
                    return fn(cnx, **kwargs)
            except Exception as err:
                raise SystemExit(str(err))

        return wrapped  # type: ignore

    if fl is None or isinstance(fl, logging.Logger):
        logger = fl
        return wrapper

    return wrapper(fl)


def add_conn_args(parser: ArgumentParser) -> None:
    "add default arguments"

    def path_pair(v: str) -> tuple[Path, Path]:
        try:
            from_pfx, to_pfx = v.split(":")
            return (Path(from_pfx), Path(to_pfx))
        except ValueError:
            pass
        raise ArgumentTypeError(f"'{v}' is not a valid value, must specify a pair of paths as'<from-path>:<to-path>'")

    g = parser.add_argument_group("connection parameters")
    g.add_argument(
        "--keyfile-pfx-map", type=path_pair, help="change private_key_file path prefix (format: <from-path>:<to-path>)"
    )
    g.add_argument("-c", "--conn", dest="connection_name", help="connection name")
    g.add_argument("--database", metavar="", help="override or set the default database")
    g.add_argument("--role", metavar="", help="override or set the default role")
    g.add_argument("--schema", metavar="", help="override or set the default schema")
    g.add_argument("--warehouse", metavar="", help="override or set the default warehouse")

    parser.add_argument(
        "--debug", dest="loglevel", action="store_const", const=logging.DEBUG, default=logging.WARNING, help=SUPPRESS
    )


def with_connection_args(doc: str | None, **kwargs: Any) -> Callable[..., Callable[..., Any]]:
    """Function decorator that instantiates and adds snowflake database connection arguments"""

    def getargs(fn: Callable[[ArgumentParser], None]) -> Callable[..., Any]:
        @wraps(fn)
        def wrapped(args: list[str] | None = None) -> Any:
            parser = ArgumentParser(description=doc, **kwargs)
            add_conn_args(parser)
            fn(parser)
            return parser.parse_args(args)

        return wrapped

    return getargs



"test getconn() method"
from sfconn import Connection, Cursor, getconn


def test_getconn() -> None:
    with getconn() as cnx, cnx.cursor() as csr:
        assert isinstance(cnx, Connection)
        assert isinstance(csr, Cursor)
        csr.execute("select current_user() as user, current_role() as role")
        assert csr.rowcount == 1



"Private key"
import os
from functools import cached_property
from pathlib import Path

import cryptography.hazmat.primitives.serialization as Serde
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey


class PrivateKey:
    "class to extract various pieces of information from private-key file"

    def __init__(
        self,
        private_key_file: Path,
        pass_phrase: str | None = None,
        pass_phrase_var: str | None = "SNOWSQL_PRIVATE_KEY_PASSPHRASE",
    ):
        if not private_key_file.is_file():
            raise FileNotFoundError(f"Invalid private Key file '{private_key_file}'")

        self.private_key_file = private_key_file
        self.pass_phrase = os.environ.get(pass_phrase_var) if pass_phrase is None and pass_phrase_var is not None else pass_phrase

    @cached_property
    def key(self) -> RSAPrivateKey:
        with self.private_key_file.open("rb") as fh:
            return Serde.load_pem_private_key(  # type: ignore
                fh.read(), password=self.pass_phrase.encode() if self.pass_phrase is not None else None, backend=default_backend()
            )

    @property
    def pri_bytes(self) -> bytes:
        return self.key.private_bytes(
            encoding=Serde.Encoding.DER, format=Serde.PrivateFormat.PKCS8, encryption_algorithm=Serde.NoEncryption()
        )

    @property
    def pub_bytes(self) -> bytes:
        return self.key.public_key().public_bytes(Serde.Encoding.DER, Serde.PublicFormat.SubjectPublicKeyInfo)



"Snowflake cursor class"
from inspect import currentframe, getframeinfo
from logging import DEBUG, getLogger
from typing import Any, Callable, Iterable, Self, Sequence, TypeAlias, TypeVar, cast

from snowflake.connector.cursor import SnowflakeCursor
from snowflake.connector.errors import ProgrammingError

logger = getLogger(__name__)
T = TypeVar("T")
Params: TypeAlias = Sequence[Any] | dict[Any, Any] | None


class Cursor(SnowflakeCursor):
    "A Cursor class that adds a few convenience methods to Snowflake provided cursor class"

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any, **kwargs: Any):
        return super().__exit__(*args, **kwargs)

    def execute_debug(self, sql: str, params: Params = None) -> Self:
        """execute a SQL statement in a debug mode by logging each SQL at DEBUG level

        Args:
            sql: SQL query text
            params: statement parameters, optional

        Returns:
            Self
        """
        if logger.getEffectiveLevel() <= DEBUG:
            fi = getframeinfo(currentframe().f_back.f_back)  # type: ignore
            logger.debug(
                "Running SQL, file: %s, line: %d, function: %s\n%s;",
                fi.filename,
                fi.lineno,
                fi.function,
                sql.replace("\t", "    "),
            )
        self.execute(sql, params=params)
        return self

    def run_query(self, rec: Callable[..., T] | type[T], sql: str, params: Params = None) -> Iterable[T]:
        """execute a SELECT statement, map rows to produce instances of type T

        Args:
            rec: must be either
                 - a Callable that accepts each column of a as a parameter in the same order and returns instance of T
                 - a class T, whose __init__ method argument names must match some column names when upper-cased.
                   Only the matching column names are passed to T.__init__() in the argument order
            sql: SELECT SQL query text
            params: optional parameters as a Sequence or a dict

        Returns:
            Iterable over instance of the type T
        """

        def mk_class(fn: Callable[..., T]) -> Iterable[T]:
            return (fn(*r) for r in cast(Iterable[tuple[Any, ...]], self.execute_debug(sql, params)))

        def get_elems() -> list[str]:
            try:
                elems = [c.upper() for c in cast(dict[str, type], rec.__init__.__annotations__) if c != "return"]
                if len(elems) == 0:
                    raise TypeError(f"'{rec}.__init__()' takes no arguments")
                return elems
            except AttributeError:
                raise TypeError(f"'{rec}.__init__()' takes no arguments")

        def get_col_pos(elems: list[str]) -> list[int]:
            try:
                cols = {d.name: e for e, d in enumerate(self.description)}
                return [cols[a] for a in elems]
            except KeyError as err:
                raise ProgrammingError(f"Column access error: {err} in SQL: {sql}")

        if isinstance(rec, type):
            elems = get_elems()
            self.execute_debug(sql)
            pos = get_col_pos(elems)
            return mk_class(lambda *r: rec(*(r[e] for e in pos)))  # type: ignore

        if callable(rec):
            return mk_class(rec)

        raise TypeError(f"'{rec}' must be either ")

    def run_query1_opt(self, rec: Callable[..., T] | type[T], sql: str, params: Params = None) -> T | None:
        """execute a SELECT statement, return the first row mapped using the provided function

        Args:
            rec: must be either
                 - a Callable that accepts each column of a as a parameter in the same order and returns instance of T
                 - a class T, whose __init__ method argument names must match some column names when upper-cased.
                   Only the matching column names are passed to T.__init__() in the argument order
            sql: SELECT SQL query text
            params: optional parameters as a Sequence or a dict

        Returns:
            Iterable over instance of the type T
        """
        return next(iter(self.run_query(rec, sql, params)), None)

    def run_query1(self, rec: Callable[..., T] | type[T], sql: str, params: Params = None) -> T:
        """execute a SELECT statement, return the first row mapped using the provided function

        Args:
            rec: must be either
                 - a Callable that accepts each column of a as a parameter in the same order and returns instance of T
                 - a class T, whose __init__ method argument names must match some column names when upper-cased.
                   Only the matching column names are passed to T.__init__() in the argument order
            sql: SELECT SQL query text
            params: optional parameters as a Sequence or a dict

        Returns:
            Iterable over instance of the type T

        Raises:
            ProgrammingError if a row is not available
        """
        row = self.run_query1_opt(rec, sql, params)
        if row is None:
            raise ProgrammingError("no data available")

        return row



"get a snowflake connection using connections.toml configuration with added convenience methods"
from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Any, Self, cast

from snowflake.connector.config_manager import CONFIG_MANAGER
from snowflake.connector.connection import SnowflakeConnection, SnowflakeCursor
from snowflake.connector.errors import Error

from .cursor import Cursor

logger = logging.getLogger(__name__)


def _parse_keyfile_pfx_map() -> tuple[Path, Path] | None:
    if (x := os.environ.get("SFCONN_KEYFILE_PFX_MAP")) is None:
        return None

    try:
        from_pfx, to_pfx = x.split(":")
        return (Path(from_pfx), Path(to_pfx))
    except ValueError:
        pass

    logger.error(f"Bad value ('{x}') for $SFCONN_KEYFILE_PFX_MAP ignored, must have a pair of paths specified as'<path>:<path>'")


_default_keyfile_pfx_map = _parse_keyfile_pfx_map()


class Connection(SnowflakeConnection):
    "A Connection class that overrides the cursor() method to return a custom Cursor class"

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any, **kwargs: Any):
        return super().__exit__(*args, **kwargs)

    def cursor(self, cursor_class: type[SnowflakeCursor] = Cursor) -> Cursor:
        return cast(Cursor, super().cursor(cursor_class))


def connection_names() -> list[str]:
    """returns names of available connections
    Returns:
        list of connection names
    """
    return list(cast(dict[str, dict[str, Any]], CONFIG_MANAGER["connections"]).keys())


def conn_opts(
    connection_name: str | None = None,
    keyfile_pfx_map: tuple[Path, Path] | None = None,
    **overrides: Any,
) -> dict[str, Any]:
    """returns connection options with overrides applied, if suplied

    Args:
        name: A connection name to be looked up from the config_file; value can be None
        keyfile_pfx_map: if specified must be a a pair of Path values specified as <from-path>:<to-path>,
           which will be used to change private_key_file path value if it starts with <from-pahd> prefix
        **overrides: A valid Snowflake python connector parameter; when not-None, will override value read from config_file

    Returns:
        dictionary containing option name and it's value

    Raises:
        *: any exceptions raised by snowflake.connector are passed through
    """
    if keyfile_pfx_map is None:
        keyfile_pfx_map = _default_keyfile_pfx_map

    def fix_keyfile_path(path: str) -> str:
        if keyfile_pfx_map is not None and (p := Path(path)).is_relative_to(keyfile_pfx_map[0]):
            return str(keyfile_pfx_map[1] / p.relative_to(keyfile_pfx_map[0]))
        return path

    connections = cast(dict[str, dict[str, Any]], CONFIG_MANAGER["connections"])
    if connection_name is None:
        connection_name = cast(str, CONFIG_MANAGER["default_connection_name"])

    if connection_name not in connections:
        raise Error(f"Invalid connection name '{connection_name}', select from [{', '.join(connections.keys())}]")

    opts = {**connections[connection_name], **{k: v for k, v in overrides.items() if v is not None}}
    if "private_key_file" in opts:
        opts["private_key_file"] = fix_keyfile_path(cast(str, opts["private_key_file"]))

    return opts


def getconn(connection_name: str | None = None, **overrides: Any) -> Connection:
    """connect to Snowflake database using named configuration

    Args
        name: A connection name to be looked up from the config_file, optional defaults to None for default connection
        **overrides: Any parameter that is valid for conn_opts() method; see conn_opts() documentation

    Returns:
        Connection object returned by Snowflake python connector
    """
    return Connection(**conn_opts(connection_name, **overrides))  # type: ignore



"test Cursor class"
import datetime as dt
from collections import namedtuple
from dataclasses import dataclass

import pytest

from sfconn import ProgrammingError, getconn

SQL = "select current_user() as user, current_date() as date"


def test_run_query_func() -> None:
    Result = namedtuple("Result", ["user", "date"])  # type: ignore

    def mkResult(x: str, y: dt.date) -> Result:
        return Result(x, y)

    with getconn() as cnx, cnx.cursor() as csr:
        csr.run_query1(mkResult, SQL)


def test_run_query_class() -> None:
    @dataclass
    class Result:
        user: str
        date: dt.date

    with getconn() as cnx, cnx.cursor() as csr:
        assert isinstance(csr.run_query1(Result, SQL), Result)


def test_run_query_class_elem_order() -> None:
    "class with elements defined in different order than SELECT"

    @dataclass
    class Result:
        date: dt.date
        user: str

    with getconn() as cnx, cnx.cursor() as csr:
        assert isinstance(csr.run_query1(Result, SQL), Result)


def test_run_query_class_fewer_elems() -> None:
    "class with elements defined with fewer elements than SELECT"

    @dataclass
    class Result:
        date: dt.date

    with getconn() as cnx, cnx.cursor() as csr:
        assert isinstance(csr.run_query1(Result, SQL), Result)


def test_run_query_class_noinit() -> None:
    class Result:
        user: str
        date: dt.date

    with pytest.raises(TypeError):
        with getconn() as cnx, cnx.cursor() as csr:
            csr.run_query1(Result, SQL)


def test_run_query_bad_columns() -> None:
    "when __init__ has arguments whose names do not match any columns"

    class Result:
        user: str
        date: dt.date

        def __init__(self, x: str, y: str) -> None:
            self.user, self.role = x, y

    with pytest.raises(ProgrammingError):
        with getconn() as cnx, cnx.cursor() as csr:
            csr.run_query1(Result, SQL)


def test_select1_norows() -> None:
    "test an exception is thrown when no rows are returned for run_query1() call"
    with pytest.raises(ProgrammingError):
        with getconn() as cnx, cnx.cursor() as csr:
            csr.run_query1(lambda *x: x, SQL + " where 1 = 0")
